/*
 * console.c
 */

#include "em_device.h"
#include "em_chip.h"

// console
#include "console.h"
// console subsystems
#include "cons_analog.h"
#include "cons_game.h"
#include "cons_led.h"
#include "cons_owner.h"
#include "cons_timedate.h"


/* shell command prototypes */
void console_cmd_help(int argc, char **argv);
void console_cmd_clear_screen(int argc, char **argv);
void console_cmd_save(int argc, char **argv);
void console_cmd_credits(int argc, char **argv);


/* shell */
ntshell_t ntshell;
text_history_t ntshell_history;
char ntshell_prompt_old[16];
static uint16_t shell_flags;

static int (*console_puts)(char *buf, int cnt);
static int (*console_gets)(char *buf, int cnt);

/* shell commands */
static const command_table_t console_cmdlist[] = {
		{"help",    "this text, '?' works too; also try 'help <cmd>'",	0x00, console_cmd_help}, 				// done
		{"?", 	    "",						                            0x00, console_cmd_help},
		{"me",      "print or set badge owner information",				0x00, console_cmd_owner_info}, 			// TODO
		{"clear",   "clears the screen",								0x00, console_cmd_clear_screen}, 		// done
		{"time",    "print the current time, or set with 'time set'",	0x00, console_cmd_time_time}, 			// done
		{"date",    "print the current date, or set with 'date set'",	0x00, console_cmd_time_date}, 			// done
		{"bclock",  "enable or disable the binary clock",				0x00, console_cmd_led_bclock}, 			// done
		{"alarm",   "submode: enable, disable, or set vibrator alarms", 0x00, console_cmd_time_alarm}, 			// done
		{"vibrate", "configure the alarm vibrator", 					0x00, console_cmd_time_alarm_vibrate},	// doing now
		{"lsens",   "print the detected light level",					0x00, console_cmd_lsens_show},
		{"save",    "saves alarms and programs to flash memory",		0x00, console_cmd_save}, 				// in progress
		{"the",     "game",												0x00, console_cmd_the_game},
		{"eye",     "submode: configure the eye rgbled programs",		0x00, console_cmd_led_eye},
		{"skull",   "submode: configure the skull border led programs",	0x00, console_cmd_led_skull},
		{"bone",    "submode: configure the bone rgbled programs",		0x00, console_cmd_led_bone},
		{"credits", "info about the whiskey pirates",					0x00, console_cmd_credits},
		{NULL, NULL, 0x00, NULL}
};
command_table_t *ctbl;
command_table_t *ctbl_prev;

/* additional */
char console_line[64];

/* hijack serial for custom text parsing */
ntshell_t hijack;


/* ntshell handlers */
static int console_read(char *buf, int cnt, void *extobj)
{
	if (console_gets != 0)
		return console_gets(buf, cnt);

	return 0;
}

int console_write(const char *buf, int cnt, void *extobj)
{
	if (console_puts != 0) {
		console_puts((char *)buf, cnt);
		return cnt;
	}

	return 0;
}

static int console_ntopt_cb(int argc, char **argv, void *extobj)
{
	// ntshell_t *ntshell = (ntshell_t *)extobj;
	char cmd[2][4];
	command_table_t *p;

	if (argc == 0) {
		return 0;
	}

	cmd[0][2] = cmd[1][2] = 0;
	strncpy(cmd[0], argv[0], 2);

	for (p = ctbl; p->cmd != NULL; p++) {
		strncpy(cmd[1], p->cmd, 2);
		if ((strcmp(cmd[0], cmd[1]) == 0) && (strlen(argv[0]) <= strlen(p->cmd))) {
			p->func(argc, argv);
			return 0;
		}
	}

	sprintf(console_line, "* '%s': ", argv[0]);
	CONS_PRINT(console_line);
	CONS_PRINT("command not found\r\n");
	return -1;
}

static int console_cb(const char *text, void *extobj)
{
	// ntshell_t *ntshell = (ntshell_t *)extobj;

	if ((shell_flags & 0x01) == 0) {
		CONS_PRINT("Whiskey Pirates 2016 KeyBadge, version 0.0.1\r\n");
		CONS_PRINT("Use 'help' or '?' for the command list.\r\n");
		shell_flags |= 0x01;
		return 0;
	}

	if (strlen(text) > 0) {
		return ntopt_parse(text, console_ntopt_cb, extobj);
	}

	return 0;
}


/* setup */
void console_init()
{
	// set up command table pointer
	ctbl = (command_table_t *)console_cmdlist;

	// set up the shell
	ntshell_init(
	      &ntshell,
	      console_read,
	      console_write,
	      console_cb,
	      &ntshell_history,
	      (void *)&hijack);

	ntshell_set_prompt(&ntshell, CONS_PROMPT_DEFAULT);
	// ntshell_execute(&ntshell);

	// set up the hijacker
	ntshell_init(
		  &hijack,
		  console_read,
		  console_write,
		  NULL,
		  0,
		  (void *)&ntshell);
	ntshell_set_prompt(&hijack, "");
}

void console_set_puts(int (*puts)(char *buf, int cnt))
{
	console_puts = puts;
}

void console_set_gets(int (*gets)(char *buf, int cnt))
{
	console_gets = gets;
}

void console_start()
{
	ntshell_execute_start(&ntshell);
}

void console_stop()
{
	shell_flags = 0;
}

void console_process()
{
	ntshell_execute_process(hijack.func_callback ? &hijack : &ntshell);
}

void console_hijack(int (*cb)(const char *text, void *extobj))
{
	if (cb) {
		// we are in the middle of processing our old shell, and we don't
		// want to show the prompt. so back it up and unset it
		// before enabling the hijack
		strncpy(ntshell_prompt_old, ntshell.prompt, 15);
		strcpy(ntshell.prompt, "");
		hijack.func_callback = cb;
		return;
	}

	// we're done, so recover our shell prompt and print it
	hijack.func_callback = 0;
	strncpy(ntshell.prompt, ntshell_prompt_old, 15);
	CONS_PRINT("\r\n");
	CONS_PRINT(ntshell.prompt);
}

/* built-in console commands */
void console_help(int argc, char **argv, command_table_t *p)
{
	int i;

	if (argc < 2) {
		// print help list
		if ((shell_flags & 0x02) == 0) {
			CONS_PRINT("Tip: You only need to use the first two characters ");
			CONS_PRINT("of a command.\r\n\r\n");
			shell_flags |= 0x02;
		}

		while (p->cmd != NULL) {
			if (strlen(p->desc)) {
				strcpy( console_line, "* ");
				strncat(console_line, p->cmd,  12);
				for (i = strlen(console_line); i <= 12; i++) {
					console_line[i] = 0x20;
				}
				console_line[i] = 0x00;
				strncat(console_line, p->desc, 77);
				strncat(console_line, "\r\n",  79);
				console_line[63] = 0;
				CONS_PRINT(console_line);
			}
			p++;
		}
	} else {
		while (p->cmd != NULL) {
			// calling the callback with argc/argv set to 0 should instruct it to print help/usage
			if (strcmp(argv[1], p->cmd) == 0) {
				p->func(0, (char **)0);
				return;
			}
			p++;
		}

		// looks like the command wasn't found
		strcpy( console_line, "* no help found for command called '");
		strncat(console_line, argv[1], 60);
		strncat(console_line, "'\r\n", 62);
		CONS_PRINT(console_line);
	}
}
void console_cmd_help(int argc, char **argv)
{
	console_help(argc, argv, (command_table_t *)console_cmdlist);
}

void console_cmd_credits(int argc, char **argv)
{
	if (!argc) {
		CONS_PRINT("Just run the command.");
		return;
	}

	console_cmd_clear_screen(1, 0);
	CONS_PRINT(" It seems you have a badge with our name on it. We are the");
	CONS_PRINT("\x1B[3;20H\x1B[01mWHISKEY PIRATE CREW\x1b[00m");
	CONS_PRINT("\x1B[5;18HBooze. Badges. Bullshit.\r\n");
}

void console_cmd_clear_screen(int argc, char **argv)
{
	if (!argc) {
		CONS_PRINT("...");
		return;
	}

	CONS_PRINT("\x1B[2J\x1B[0;0H");
}

void console_settings_save()
{
	CONS_PRINT("Saving settings...");
	CONS_PRINT("done.\r\n");
}

void console_cmd_save(int argc, char **argv)
{
	if (!argc) {
		CONS_PRINT("Saves settings to flash. Run the 'save' command for arguments.");
	}

	if (argc == 1) {
		CONS_PRINT("Usage: save <all/setup or settings/alarm/led)\r\n");
	}

	if (argc == 2) {
		strtolower((unsigned char *)argv[1]);
		if (strcmp(argv[1], "all") == 0) {
			console_settings_save();
			console_time_alarm_save();
		}
		if (strcmp(argv[1], "alarm") == 0) {
			console_time_alarm_save();
		}
		if (strcmp(argv[1], "led") == 0) {

		}
		if (argv[1][0] == 's' && argv[1][1] == 'e') {
			console_settings_save();
		}
	}
}
