/**
 * lightsense.c
 */

#include "em_device.h"
#include "em_chip.h"

#include "tools.h"

#include "adc.h"
#include "lightsense.h"

#include "display/process.h"


static uint8_t  lsens_state;
static uint16_t lsens_wait;

uint16_t lsens_count;
uint8_t  lsens_gain = 128; 		// starting default = a bit dim


const uint8_t lsens_ref[] = {
	255, 		// dark / dusk
	 40, 		// low
	 24, 		// normal
	  2, 		// high
	  0 		// bright
};


/* hardware */
void lsens_led_output(uint8_t out)
{
	// set cathode as low output
	GPIO_PinModeSet(LSENS_K_PORT, LSENS_K_PIN, gpioModePushPull, 0);
	// set anode per out parameter
	GPIO_PinModeSet(LSENS_A_PORT, LSENS_A_PIN, gpioModePushPull, out);
}

void lsens_led_config(GPIO_Mode_TypeDef mode, uint8_t out)
{
	// set cathode as appropriate
	GPIO_PinModeSet(LSENS_K_PORT, LSENS_K_PIN, mode, out);
	// set anode low
	GPIO_PinModeSet(LSENS_A_PORT, LSENS_A_PIN, gpioModePushPull, 0);
}

void lsens_adc_input()
{
	adc_set_chan(ADC_INPUT_LIGHTSENS);
}

/* light sensing */
static inline void lsens_process(uint16_t adc_result)
{
	int8_t adj = 0;

	// do normal
	lsens_count = adc_result;

	if (adc_result > 3800) {
		adj = 4;
	} else if (adc_result > 3500) {
		adj = 1;
	} else if (adc_result < 1700) {
		adj = -4;
	} else if (adc_result < 2700) {
		adj = -1;
	}

	if (adj > 0) {
		if (lsens_gain < (239 - adj)) {
			lsens_gain += adj;
		} else {
			lsens_gain = 239;
		}
	} else if (adj < 0) {
		if (lsens_gain > (0 - adj)) {
			lsens_gain += adj;
		} else {
			lsens_gain = 0;
		}
	}
}

void lsens_run(uint8_t start)
{
	volatile int wait;

	if (start) {
		lsens_state = LSENS_STATE_START;
	}

	switch (lsens_state) {
		case LSENS_STATE_IDLE: {
			break;
		}
		case LSENS_STATE_START: {
			// get waiting value
			lsens_wait = lsens_gain * 2;

			// set up ADC
			lsens_adc_input();

			// short LED, then reverse bias and wait a tiny while
			lsens_led_output(0);
			lsens_led_config(gpioModePushPull, 1);
			for (wait = 10; wait; wait--);

			// LED is charged, set as disabled for ADC
			lsens_led_config(gpioModeDisabled, 0);

			// update state
			lsens_state = LSENS_STATE_WAIT;
			break;
		}
		case LSENS_STATE_WAIT: {
			if (lsens_wait) {
				lsens_wait--;
				break;
			}
		}
		// no break
		case LSENS_STATE_START_ADC: {
			// see how many photons hit
			adc_start();
			lsens_state = LSENS_STATE_ACQUIRE;
			break;
		}
		case LSENS_STATE_ACQUIRE: {
			// more than enough time has passed; get value from ADC
			lsens_process(adc_getresult(ADC_INPUT_LIGHTSENS));
			lsens_state = LSENS_STATE_IDLE;
			break;
		}

		default: {
			// start new process
			lsens_state = LSENS_STATE_IDLE;
		}
	}
}

/* helpers */
uint16_t light_scale(uint8_t val, uint16_t src_max, uint16_t dst_max)
{
	int32_t  i;
	uint16_t w = 0;

	for (i = 0; i < 4; i++) {
		if (lsens_gain < lsens_ref[i]) {
			w = scale16ufast(val, src_max, dst_max);
			break;
		}
	}

	return w;
}
