/**
 * userio.c
 */

/* efm32 related */
#include "em_device.h"
#include "em_chip.h"
#include "em_emu.h"
#include "em_gpio.h"

/* hardware interfaces */
#include "leuart.h"
#include "shiftreg.h"
#include "../display/led.h"
#include "../usb/usb.h"

/* console-related, for setting up comms settings */
#include "console/console.h"
#include "comms/keyuart/skeltal.h"
#include "comms/usb/basic_cdc.h"

/* common */
#include "InitDevice.h"
#include "tools.h"

/* this */
#include "userio.h"


/* user io */
static uint8_t userio_state;
static BtnState userio_btn;


/* it begins */
static uint8_t userio_read_state()
{
	uint8_t ret;

	ret  = GPIO_PinInGet(UIN_TOG_SLEEP_PORT, UIN_TOG_SLEEP_PIN);
	ret |= GPIO_PinInGet(UIN_TOG_MODE_PORT, UIN_TOG_MODE_PIN) << 1;

	return ret;
}


void userio_init()
{
	// configure IO
	// this has been done for us already...

	// configure buttons
	userio_btn.state = UIN_RELEASE;
	userio_btn.debounce = UIN_DEBOUNCE_CYCLES + 1;

	// check state now
	userio_state = userio_read_state();

	// configure interrupts for toggle switches
	GPIO_IntConfig(UIN_TOG_MODE_PORT, UIN_TOG_MODE_PIN, 1, 1, 1);
	GPIO_IntConfig(UIN_TOG_SLEEP_PORT, UIN_TOG_SLEEP_PIN, 1, 1, 1);
	GPIO->INSENSE = GPIO_INSENSE_INT;

	NVIC_ClearPendingIRQ(GPIO_EVEN_IRQn);
	NVIC_EnableIRQ(GPIO_EVEN_IRQn);

	NVIC_ClearPendingIRQ(GPIO_ODD_IRQn);
	NVIC_EnableIRQ(GPIO_ODD_IRQn);

	// do necessary actions now
	userio_cb(UIN_TOG_SLEEP);
}

uint8_t userio_update(uint8_t alternate)
{
	uint8_t read;

	read = GPIO_PinInGet(UIN_BTN_ANY_PORT, UIN_BTN_ANY_PIN);

	if (alternate) {
		return read ? 1 : 0;
	}

	// not pushed
	if (read) {
		// are we in release state?
		if (userio_btn.state & UIN_RELEASE) {
			// continue releasing
			if (userio_btn.debounce < UIN_DEBOUNCE_CYCLES) {
				userio_btn.debounce++;
			}
			// done releasing
			else if (userio_btn.debounce == UIN_DEBOUNCE_CYCLES) {
				// clear all other states
				userio_btn.state = UIN_RELEASE;
				// set so we won't process anymore
				userio_btn.debounce++;
				// and callback user function
				if (userio_btn.cb_release) userio_btn.cb_release();
			}
		} else {
			// start releasing
			userio_btn.state |= UIN_RELEASE;
			userio_btn.debounce = 0;
		}
	}
	// pushed
	else {
		// are we in push state?
		if (userio_btn.state & UIN_PUSH) {
			// have we held the button yet?
			if (userio_btn.state & UIN_HOLD) {
				// are we continuing to hold the button?
				if (userio_btn.state & UIN_HOLD_REPEAT) {
					// continue holding
					if (userio_btn.debounce < userio_btn.hold_cycles) {
						userio_btn.debounce++;
					}
					// done holding
					else if (userio_btn.debounce == userio_btn.hold_cycles) {
						// reset debounce and callback user function
						userio_btn.debounce = 0;
						if (userio_btn.cb_hold) userio_btn.cb_hold(1);
					}
				}
				// no, this is the first hold
				else {
					// continue holding
					if (userio_btn.debounce < userio_btn.hold_cycles) {
						userio_btn.debounce++;
					}
					// done holding
					else if (userio_btn.debounce == userio_btn.hold_cycles) {
						userio_btn.debounce++;
						// set repeat flag and debounce if repeat is enabled
						if (userio_btn.option & UIN_OPT_HOLD_REPEAT) {
							userio_btn.state |= UIN_HOLD_REPEAT;
							userio_btn.debounce = 0;
						}
						// and callback user function
						if (userio_btn.cb_hold) userio_btn.cb_hold(0);
					}
				}
			}
			// we haven't pushed the button yet, debounce
			else {
				// continue pushing
				if (userio_btn.debounce < UIN_DEBOUNCE_CYCLES) {
					userio_btn.debounce++;
				}
				// done pushing
				else if (userio_btn.debounce == UIN_DEBOUNCE_CYCLES) {
					// clear all other states, prepare to hold
					userio_btn.state = UIN_PUSH | UIN_HOLD;
					// reset debounce and callback user function
					userio_btn.debounce = 0;
					if (userio_btn.cb_push) userio_btn.cb_push();
				}
			}
		} else {
			// start pushing
			userio_btn.state |= UIN_PUSH;
			userio_btn.state &= ~(UIN_HOLD | UIN_HOLD_REPEAT);
			userio_btn.debounce = 0;
		}
	}

	return 0;
}

void userio_cb(uint8_t swidx)
{
	switch (swidx) {
		case UIN_TOG_SLEEP: {
			if (userio_state & (1 << UIN_TOG_SLEEP)) {
				// --- high == switch down == sleep
				// disable systick interrupt
				NVIC_DisableIRQ(SysTick_IRQn);
				// disable LEDs
				led_stop();
			} else {
				// --- low == switch up = run
				// set everything back up to default
				enter_DefaultMode_from_RESET();
				// set LED states back to default starting values
				led_init();
				// restore clock interrupt
				systick_isr_init();
			}
			// restore mode
		}
		// no break (fixes eclipse warning)
		case UIN_TOG_MODE: {
			if (userio_state & (1 << UIN_TOG_SLEEP)) {
				// we should be asleep so just stop all comms
				usb_stop();
				leuart_stop();
				return;
			}

			if (userio_state & (1 << UIN_TOG_MODE)) {
				// --- high == switch down == usb

				// set peripherals
				leuart_stop();
				usb_init();
				// set console handler functions
				console_set_gets(usb_cdc_gets);
				console_set_puts(usb_cdc_puts);
				// disable power pin
				sr_dat[0] &= ~SR_POWER_MASK;
			} else {
				// --- low == switch up == uart
				// set peripherals
				usb_stop();
				leuart_init();
				// set console handler functions
				console_set_gets(keyuart_gets);
				console_set_puts(keyuart_puts);
				// enable power pin
				sr_dat[0] |= SR_POWER_MASK;
			}
			sr_update();
			return;
		}
	}
}

void userio_sleep_enter()
{
	if (userio_state & (1 << UIN_TOG_SLEEP)) {
		EMU_EnterEM2(1);
	} else {
		EMU_EnterEM1();
	}
}


/* isr */
void GPIO_IRQHandler()
{
	uint32_t intr = GPIO_IntGet();
	GPIO_IntClear(intr);

	userio_state = userio_read_state();

	if (intr & (1 << UIN_TOG_SLEEP_PIN)) {
		userio_cb(UIN_TOG_SLEEP);
	}
	if (intr & (1 << UIN_TOG_MODE_PIN)) {
		userio_cb(UIN_TOG_MODE);
	}
}

void GPIO_EVEN_IRQHandler()
{
	GPIO_IRQHandler();
}

void GPIO_ODD_IRQHandler()
{
	GPIO_IRQHandler();
}
