/**
 * rtc/alarm.c
 */

#include "em_device.h"
#include "em_chip.h"
#include "em_cmu.h"
#include "em_rtc.h"

#include "time.h"
#include "alarm.h"

// nvram
#include "hw/spiflash.h"
#include "hw/spiflash_page.h"


/* vars */
static RTCAlarm alarm[RTC_ALARM_COUNT];
static char alarm_modetext[6];

static uint8_t alarm_lastidx;


/* alarm handling */
RTCAlarm * alarm_getptr(int idx)
{
	return &alarm[idx];
}

void alarm_delete(int idx)
{
	int i;

	// move ahead all other alarms
	for (i = idx; i < (RTC_ALARM_COUNT - 1); i++) {
		alarm[i] = alarm[i + 1];
	}

	// and clear the last alarm
	alarm[RTC_ALARM_COUNT - 1].modeflags &= ~RTC_ALARM_MODE_CREATED;

	// also clear the last fired alarm
	alarm_lastidx = 0;
}

void alarm_enable(int idx, uint8_t enable)
{
	if (enable) {
		alarm[idx].modeflags |=  RTC_ALARM_MODE_ENABLE;
	} else {
		alarm[idx].modeflags &= ~RTC_ALARM_MODE_ENABLE;

		// also clear the last fired alarm
		alarm_lastidx = 0;
	}
}

uint8_t alarm_check()
{
	int i;
	uint8_t modeflags;

	if (!rtc_is_clock_set()) {
		return 0;
	}
	for (i = 0; i < RTC_ALARM_COUNT; i++) {
		modeflags = alarm[i].modeflags;
		if (modeflags & RTC_ALARM_MODE_ENABLE) {
			if (alarm[i].hour == time_hour && alarm[i].minute == time_minute) {
				// time match, check mode
				if (modeflags & RTC_ALARM_MODE_ONESHOT) {
					alarm[i].modeflags &= ~RTC_ALARM_MODE_ENABLE;
					alarm_lastidx = i;
					return i + 1;
				}

				if (modeflags & RTC_ALARM_MODE_REPEAT) {
					alarm_lastidx = i;
					return i + 1;
				}

				// must be individual day/month, so check that now.
				if (alarm[i].month == time_month && alarm[i].day == time_day) {
					// oh shit
					alarm_lastidx = i;
					return i + 1;
				}
			}
		}
	}

	return 0;
}

char * alarm_get_modetext(int idx)
{
	if (alarm[idx].modeflags & RTC_ALARM_MODE_REPEAT) {
		return "Every";
	}
	if (alarm[idx].modeflags & RTC_ALARM_MODE_ONESHOT) {
		return "Once,";
	}

	int m;
	m = alarm[idx].month;
	if (!m) { m = 1; }

	sprintf(alarm_modetext, "%s%02d", month[m - 1], alarm[idx].day);
	return alarm_modetext;
}

uint8_t alarm_save_nv()
{
	// erase flash page
	spiflash_erasepage((FLASH_PAGESIZE * FLASH_PAGE_ALARM), 4);
	while (spiflash_is_busy());

	// store
	return spiflash_writebuf(&alarm, (FLASH_PAGESIZE * FLASH_PAGE_ALARM), sizeof(alarm), 0);
}

uint8_t alarm_restore_nv()
{
	// restore
	return spiflash_readbuf(&alarm, (FLASH_PAGESIZE * FLASH_PAGE_ALARM), sizeof(alarm), 0);
}
